import json, copy

                ######################## Methods to store program config in JSON ########
                ######################## Config can be incrementable, see mergeJSON #####
                ########################### JSON CONFIG & STUFF #########################
n_JSON_Indent = 4
s_AddKey = "ADD_KEY::"          ## The prefix for keys to be added in new backend definitions
                                ## The prefix itself is not added
                                ## TODO: remove in lvalue
s_CommentKey = "///COMMENT"     ## The key for comments in JSON config dictionaries

## Returns a merge of j2 into j1.
def mergeJSON( j1, j2 ):
    jNew = None
    if dict==type(j1) and dict==type(j2):
        jNew = copy.deepcopy( j1, {} )
        for key in j2:
            if key in j1:
                jNew[key] = mergeJSON( j1[key], j2[key] )
            else:
                assert key.startswith( s_AddKey ), "Adding key %s: does not start with the prefix %s, j1: %s" % (key, s_AddKey, j1.__str__())
                jNew[key[len(s_AddKey):]] = j2[key]
    elif list==type(j1) and list==type(j2):
        jNew = []
        for i in range( 0, len(j2) ):     ## Can assign a shorter list this way
            if i<len(j1):
                jNew.append( mergeJSON( j1[i], j2[i] ) )
            else:
                jNew.append( j2[i] )
    elif type(j1)==type(j2):
        jNew = j2
    else:
        print( "\n\n ERROR: Could not merge JSON objects:" );
        print( json.dumps( j1 ) )
        print( "\n -- TO BE MERGED BY --\n" )
        print( json.dumps( j2 ) )
        sys.quit()
    return jNew




from pathlib import Path
from minizinc import default_driver, Driver
import subprocess
import json


def _run_test(model_file):
    p = subprocess.run(
        [default_driver._executable, model_file, "--json-stream", "-a"],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p.check_returncode()
    result = [json.loads(l) for l in p.stdout.splitlines(False)]
    messages = [m for m in result if m["type"] in ["trace", "solution"]]
    assert len(messages) == 4
    trace = messages[0]
    assert trace["type"] == "trace"
    assert trace["message"] == {"url": "foo.html", "userData": {"n": 4}}
    assert messages[1]["type"] == "solution"
    assert messages[2]["type"] == "solution"
    assert messages[3]["type"] == "solution"
    expected = [{"x": 3, "y": 2}, {"x": 2, "y": 3}, {"x": 3, "y": 3}]
    actual = [s["output"]["mzn_vis_0"] for s in messages[1:]]
    assert len(actual) == len(expected)
    for sol in actual:
        assert sol in expected
    assert actual == expected


def test_vis_ann():
    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    _run_test(here / "test_vis_ann.mzn")


def test_vis_custom():
    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    _run_test(here / "test_vis_custom.mzn")



from pathlib import Path
import subprocess
import json


def test_array_dzn_output():
    from minizinc import default_driver, Driver

    expected = {
        "type": "solution",
        "output": {"dzn": "y = array3d('a b'..'a b', 'a b'..'a b', c..c, [1]);\n"},
        "sections": ["dzn"],
    }

    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    model_file = here / "test_array_dzn_output.mzn"

    p = subprocess.run(
        [
            default_driver._executable,
            model_file,
            "--json-stream",
            "--solver",
            "gecode",
        ],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert p.returncode == 0
    actual = [json.loads(l) for l in p.stdout.splitlines(False)]
    assert expected in actual



from pathlib import Path
from minizinc import default_driver, Driver
import subprocess
import json


def test_command_line_error():
    assert isinstance(default_driver, Driver)
    p = subprocess.run(
        [default_driver._executable, "--made-up-argument", "--json-stream"],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert p.returncode == 1
    messages = [json.loads(l) for l in p.stdout.splitlines(False)]
    errors = [m for m in messages if m["type"] == "error"]
    assert len(errors) == 1
    assert errors[0]["what"] == "argument parsing error"



from pathlib import Path
from minizinc import default_driver, Driver
import subprocess
import json


def test_json_stream_trace_stdout():
    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    model_file = here / "test_json_stream_trace_stdout.mzn"
    p = subprocess.run(
        [default_driver._executable, model_file, "--json-stream"],
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert p.returncode == 0
    messages = [json.loads(l) for l in p.stdout.splitlines(False)]
    traces = [m for m in messages if m["type"] == "trace"]
    assert len(traces) == 1
    assert traces[0] == {"type": "trace", "section": "default", "message": "trace"}
    solution = next(m for m in messages if m["type"] == "solution")
    expected = {
        "type": "solution",
        "output": {"default": "output", "raw": "output"},
        "sections": ["default", "raw"],
    }
    assert solution == expected



#!/usr/bin/env python3

import argparse
import math
import numpy as np
import pandas as pd
import sys
import html

from tabulate import tabulate


def stat_change(row):
    # Create [difference, percent_change, modified_percent_change]
    if row.value_old == row.value_new:
        return pd.Series([0, 0, 0])
    try:
        old = float(row.value_old)
        new = float(row.value_new)
    except ValueError:
        return pd.Series([np.NaN, np.NaN, np.NaN])

    if old == 0:
        return pd.Series([new, np.NaN, np.NaN])

    diff = new - old
    change = diff / old
    modified = change

    if row.statistic == "flatTime":
        # Reduce importance of tiny times
        t_old = math.ceil(old * 2)
        t_new = math.ceil(new * 2)
        t_change = (t_new - t_old) / t_old
        if abs(t_change) < abs(change):
            modified = t_change

    return pd.Series([diff, change, modified])


def summary_delta(row):
    # Create [difference, percentage_change]
    if row.total_old == row.total_new:
        return pd.Series([0, 0])
    if row.total_old == 0:
        return pd.Series([row.total_new, np.NaN])

    return pd.Series(
        [row.total_new - row.total_old, (row.total_new - row.total_old) / row.total_old]
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Compare flattening benchmark runs.")
    parser.add_argument(
        "baseline",
        type=str,
        help="baseline statistics CSV file",
    )
    parser.add_argument(
        "observed",
        type=str,
        help="new statistics CSV file",
    )
    parser.add_argument(
        "output",
        type=str,
        help="output HTML file",
    )
    parser.add_argument(
        "--threshold",
        default=0.1,
        type=float,
        help="threshold for showing changes",
    )

    args = parser.parse_args()

    df1 = pd.read_csv(args.baseline)
    df2 = pd.read_csv(args.observed)
    t = args.threshold

    df = pd.merge(
        df1, df2, on=["model", "data_file", "statistic"], suffixes=["_old", "_new"]
    )

    # Per instance changed statistics
    df[["difference", "change", "modified"]] = df.apply(stat_change, axis=1)
    df_c = (
        df.loc[
            (df.model == "")
            | df.modified.isna()
            | (df.modified >= t)
            | (df.modified <= -t)
        ]
        .drop(["modified"], axis=1)
        .sort_values(by=["model", "data_file", "statistic"])
    )

    # Summary statistics
    df["total_old"] = pd.to_numeric(df.value_old, errors="coerce")
    df["total_new"] = pd.to_numeric(df.value_new, errors="coerce")
    df_s = (
        df.drop(["difference", "change", "modified"], axis=1)
        .groupby(by=["statistic"])
        .sum(numeric_only=True)
        .reset_index(level=0)
        .sort_values(by="statistic")
    )
    df_s[["difference", "change"]] = df_s.apply(summary_delta, axis=1)

    print(tabulate(df_s, headers="keys", tablefmt="pretty"))

    with open(args.output, "w", encoding="utf-8") as fp:
        summary = df_s.to_html(
            index=False,
            formatters={
                "total_old": "{:,}".format,
                "total_new": "{:,}".format,
                "difference": "{:,}".format,
                "change": "{:+,.2%}".format,
            },
        )
        details = df_c.to_html(
            index=False,
            formatters={
                # "value_old": "{:,}".format,
                # "value_new": "{:,}".format,
                "difference": "{:,}".format,
                "change": "{:+,.2%}".format,
            },
        )
        fp.write(
            f"""<!DOCTYPE html>
            <html>
                <head>
                    <meta charset="utf-8">
                    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/jquery.tablesorter/2.31.3/css/theme.blue.min.css"/>
                    <script src="https://cdnjs.cloudflare.com/ajax/libs/jquery/3.6.0/jquery.slim.js"></script>
                    <script src="https://cdnjs.cloudflare.com/ajax/libs/jquery.tablesorter/2.31.3/js/jquery.tablesorter.min.js"></script>
                    <script src="https://cdnjs.cloudflare.com/ajax/libs/jquery.tablesorter/2.31.3/js/jquery.tablesorter.widgets.min.js"></script>
                    <style>
                        @import url('https://fonts.googleapis.com/css2?family=Open+Sans&display=swap');
                        html, body {{
                            font-family: 'Open Sans', sans-serif;
                        }}
                    </style>
                </head>
                <body>
                    <h2>Flattening benchmarks</h2>
                    <p>Generated with <pre><code>{html.escape(" ".join(sys.argv))}</code></pre></p>
                    <h3>Summary</h3>
                    {summary}
                    <details>
                        <summary>Per-instance significant changes</summary>
                        {details}
                    </details>
                    <script>
                        $("table").tablesorter({{
                            theme: 'blue',
                            widgets: ["zebra", "filter"]
                        }});
                    </script>
                </body>
            </html>    
            """
        )



import json, sys

                ########################### LOGFILES #########################
sChunkSep = '---'           ## A single line like this separates chunks in a logfile

##############################################################################################
###################### Logfile I/O #####################
##############################################################################################
## PROBLEM: JSON does not allow appending.
## WORKAROUND: separate [JSON] chunks by '---', inspired by https://github.com/pvorb/jsml.
## Writing an arbitrary chunk: ensure end-of-line before separator
def writeLogChunk( wf, ch ):
    ss = ch.__str__()
    wf.write( ss )
    if 0<len(ss) and not ss.endswith( '\n' ):
        wf.write( '\n' )
    wf.write( sChunkSep + '\n' )
    
## Reading an arbitrary log chunk as a string
def readLogChunk( rf ):
    ch = ''
    while True:
        line = rf.readline()
        if 0==len(line) or sChunkSep==line.strip():
            return ch
        ch += line
        
## Would return empty from empty chunk even when the original file continues
def readLogJSONChunk( rf ):
    chJ = None
    ch = readLogChunk( rf )
    if 0<len(ch):
        try:
            chJ = json.loads( ch )
        except:
            print("\n   WARNING: failed to parse JSON chunk '", ch,
                  "' in file pos ", rf.tell(), ".  SYSINFO:  ", sys.exc_info()[0]
                  , sep='')
            raise
    return chJ
            



from pathlib import Path
from minizinc import Instance, Model, Solver, Status, default_driver, Driver
import json
from minizinc_testing.utils import temporary_file


def test_output_checker():
    here = Path(__file__).resolve().parent
    assert isinstance(default_driver, Driver)
    model_file = here / "test_output_checker.mzn"
    checker_file = here / "test_output_checker.mzc.mzn"
    p = default_driver._run([model_file, checker_file, "--output-mode", "checker"])
    messages = [json.loads(l) for l in p.stdout.splitlines(False)]
    solutions = [m for m in messages if m["type"] == "solution"]
    assert len(solutions) > 0
    for s in solutions:
        model = Model(checker_file)
        instance = Instance(Solver.lookup("gecode"), model)
        with temporary_file(
            s["output"]["dzn"], suffix=".dzn"
        ) as dzn, instance.branch() as child:
            child.add_file(dzn)
            solution = child.solve()
            assert solution.status in [
                Status.ALL_SOLUTIONS,
                Status.OPTIMAL_SOLUTION,
                Status.SATISFIED,
            ]



import os, copy, io
from collections import OrderedDict

def try_float( sVal, dft=None ):
    res = dft
    try:
        res = float( sVal )
    except:
        pass
    return res

def mergeDict( dict1, dict2 ):
    for key in dict2:
        dict1[ key ] = dict2[ key ]

def addMapValues( dict1, dict2 ):
    for key in dict2:
        if key in dict1:
            dict1[key] += dict2[key]
        else:
            dict1[key] = dict2[key]
            
class MatrixByStr:          ### Matrix with string keys
    def __init__( self,
            rows, cols,          ## each a list of tuples: (short, long) row/column names
            defval = 0 ):
        self.rows, self.cols = rows, cols
        self.mbs = OrderedDict( (r[0],
            OrderedDict( (c[0], copy.deepcopy(defval)) for c in cols )) for r in rows )

    def __getitem__( self, k ):
        return self.mbs[ k[0] ][ k[1] ]

    def __setitem__( self, k, v ):
        self.mbs[ k[0] ][ k[1] ] = v

    def stringify2D( self ):           ## compact 2D print
        return MyTab().tabulate( [
            [key] + list(row.values())        ## merge each row's key into the row
              for key,row in self.mbs.items() ],
            [""] + [ col[0] for col in self.cols ] )

    ## assumes each item is a list
    def stringifyLists( self, title, rowIntro, colIntro ):
        resIO = io.StringIO()
        print( '#'*50, file=resIO )
        print( '#'*50, file=resIO )
        print( '#'*50, file=resIO )
        print( title, file=resIO )
        print( '#'*50, file=resIO )
        print( '#'*50, file=resIO )
        print( '#'*50, file=resIO )
        print( file=resIO )

        for kr, r in self.mbs.items():
            print( '#'*50, file=resIO )
            print( '#'*50, file=resIO )
            print( rowIntro, kr, file=resIO )
            print( '#'*50, file=resIO )
            print( '#'*50, file=resIO )
            print( file=resIO )

            for kc, c in r.items():
                print( '#'*50, file=resIO )
                print( rowIntro, kr, file=resIO )
                print( colIntro, "    ", kc, "  / ", len(c), file=resIO )
                print( '#'*50, file=resIO )

                for el in c:
                    print( el, file=resIO )
                print( file=resIO )

        return resIO.getvalue()

class MyTab:
    def __init__( self ):
        self.sColSep='  '
      
    ## Receive table row-wise (matr)
    ## hdr is a list of column headers
    def tabulate( self, matr, hdr ):
        res = ""
        ## Compute max width
        nWMax = []
        for iC in range( len( hdr ) ):
            nWMax.append( len( str( hdr[ iC ] ) ) )
            for iR in range( len( matr ) ):
                if iC<len( matr[ iR ] ):
                    if nWMax[ iC ]<len( str( matr[iR][iC] ) ):
                        nWMax[ iC ] = len( str( matr[iR][iC] ) )
        ## Printing
        res =( ("{0:<"+str(nWMax[0])+'}').format(hdr[0]) + self.sColSep )
        for iC in range( 1, len( hdr ) ):
            res += ( ("{0:"+str(nWMax[iC])+'}').format(hdr[iC])
                  + ( self.sColSep if iC+1<len(hdr) else '\n' ) )
        for iR in range( len( matr ) ):
            res += ( ("{0:<"+str(nWMax[0])+'}').format(matr[iR][0] if 0<len( matr[ iR ] ) else '-') + self.sColSep )
            for iC in range( 1, len( matr[iR] ) ):
                res += ( ("{0:"+str(nWMax[iC])+'}').format(matr[iR][iC] if iC<len( matr[ iR ] ) else '-')
                       + ( self.sColSep if iC+1<len(hdr) else '\n' ) )
        return res

## change string to be abke to become a filename
def flnfy( sStr ):
    keepcharacters = ('-','_')
    return "".join(c if c.isalnum() or c in keepcharacters else 'I' for c in sStr).strip()

def makeDirname( sFln ):
    sDir = os.path.dirname( sFln )
    if 0<len(sDir):
        try:
            os.makedirs( sDir, exist_ok=True)
        except:
            print( "Failed to create dir '", sDir, "'.", sep='' )
            return False
    return True

def openFile_autoDir( sFln, sMode ):
    makeDirname( sFln )
    return open( sFln, sMode )

## stringify a container, inserting a given prefix before each element
def strNL( prf, cnt ):
    return prf + prf.join( [ str( el ) for el in cnt ] )





from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile


@contextmanager
def temporary_file(contents: str,  **kwargs) -> Path:
    t = NamedTemporaryFile("w", encoding="utf-8", delete=False, **kwargs)
    p = Path(t.name)
    try:
        t.write(contents)
        t.close()
        yield p
    finally:
        p.unlink(missing_ok=True)



#!/usr/bin/env python3

from pathlib import Path
from subprocess import run
from typing import List, Tuple
import argparse
import asyncio
import csv
import json
import os
import shlex
import time


async def flatten(
    instance: List[Path],
    minizinc: str,
    prefix_args: List[str],
    suffix_args: List[str],
):
    cmd = []
    cmd.extend(prefix_args)
    cmd.extend(
        [
            minizinc,
            "--solver",
            "org.minizinc.mzn-fzn",
            "-c",
            "-s",
            "-v",
        ]
    )
    cmd.extend(suffix_args)
    cmd.extend(instance)
    p = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    stdout, stderr = await p.communicate()
    return p.returncode, stdout.decode(), stderr.decode()


async def process(
    semaphore: asyncio.Semaphore,
    count: Tuple[int, int],
    instance: List[Path],
    minizinc: str,
    prefix_args: List[str],
    suffix_args: List[str],
    root: Path,
    stats,
):
    files = [p.relative_to(root).as_posix() for p in instance]
    name = " with ".join(files)
    i, n = count
    async with semaphore:
        print(f"Flattening {name} [{i}/{n}]...")
        exit_code, stdout, stderr = await flatten(
            instance, minizinc, prefix_args, suffix_args
        )
    while len(files) < 2:
        files.append("")
    for line in stdout.splitlines():
        l = line.strip()
        if l.startswith("%%%mzn-stat: "):
            [key, val] = l[13:].split("=")
            try:
                val = json.loads(val)
            except json.JSONDecodeError:
                pass
            stats.writerow(files + [key, val])

    if exit_code == 0:
        print(f"Flattening {name} succeeded.")
        return True
    else:
        print(f"Flattening {name} failed.")
        print(stderr)
        return False


async def main(args):
    minizinc = args.minizinc
    instances = []
    version = run([minizinc, "--version"], capture_output=True, encoding="utf-8").stdout
    print("Flattening benchmarking tool\n")
    print(version)
    root = Path(args.instances)
    if root.is_dir():
        for mzn in root.rglob("*.mzn"):
            dzns = list(mzn.parent.rglob("*.dzn"))
            if len(dzns) > 0:
                for dzn in dzns:
                    instances.append([mzn, dzn])
            else:
                instances.append([mzn])
    elif root.is_file():
        with root.open("r", newline="") as fp:
            for row in csv.reader(fp):
                instances.append([Path(root.parent) / f for f in row if len(f) > 0])
        root = root.parent
    else:
        raise ValueError(f"Failed to collect instances from '{root}'")
    n = len(instances)
    max_workers = n
    if args.parallel == 0:
        max_workers = os.cpu_count()
    elif args.parallel > 0:
        max_workers = args.parallel
    semaphore = asyncio.Semaphore(max_workers)
    print(f"{n} instances collected.")
    print(f"Flattening with {max_workers} workers.\n")
    prefix_args = shlex.split(args.prefix_args)
    suffix_args = shlex.split(args.args)
    with open(args.output, "w", newline="") as csvfile:
        stats = csv.writer(csvfile)
        stats.writerow(["model", "data_file", "statistic", "value"])
        stats.writerow(["", "", "_minizincVersion", version])
        stats.writerow(["", "", "_minizincCommand", minizinc])
        stats.writerow(["", "", "_minizincPrefixArgs", prefix_args])
        stats.writerow(["", "", "_minizincSuffixArgs", suffix_args])
        results = await asyncio.gather(
            *[
                process(
                    semaphore,
                    (i, n),
                    instance,
                    minizinc,
                    prefix_args,
                    suffix_args,
                    root,
                    stats,
                )
                for i, instance in enumerate(instances)
            ]
        )
        fails = results.count(False)
        stats.writerow(["", "", "_flatFailures", fails])
        print(f"{fails}/{n} instances failed to flatten.")

    try:
        # Print summary if dependencies are available
        import pandas as pd
        from tabulate import tabulate

        df = pd.read_csv(args.output)
        df.value = pd.to_numeric(df.value, errors="coerce")
        df_s = (
            df.groupby(by=["statistic"])
            .sum(numeric_only=True)
            .reset_index(level=0)
            .sort_values(by="statistic")
        )
        print(tabulate(df_s, headers=["statistic", "total"], tablefmt="pretty"))
    except ImportError:
        pass


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run flattening benchmark.")
    parser.add_argument(
        "instances",
        type=str,
        help="instances directory",
    )
    parser.add_argument(
        "output",
        type=str,
        help="output CSV file",
    )
    parser.add_argument(
        "--minizinc",
        default="minizinc",
        type=str,
        help="Path to the minizinc binary to use",
    )
    parser.add_argument(
        "--prefix-args",
        default="",
        type=str,
        help="Command line options to add to beginning of minizinc call",
    )
    parser.add_argument(
        "--args",
        default="",
        type=str,
        help="Command line options to add to end of minizinc call",
    )
    parser.add_argument(
        "--parallel",
        default=0,
        type=int,
        help="Parallel instances to flatten",
    )

    start = time.time()
    asyncio.run(main(parser.parse_args()))
    print(f"Done in {time.time() - start}s.")
